# -*- coding: utf-8 -*-
""" 
    msiem constants
"""
BASE_URL = 'https://{}/rs/esm/'
BASE_URL_PRIV = 'https://{}/ess'

ASYNC_MAX_WORKERS=10
GENERAL_POST_TIMEOUT=30

CONFIG_FILE_NAME='.msiem/conf.ini'

DEFAULT_CONFIG_FILE='''
    ; This file should be located in your securely in your path since it 
    ; has credentials.
    ; 
    ; For Windows:  %APPDATA%\\\\'''+CONFIG_FILE_NAME+'''
    ; For Mac :     $HOME/'''+CONFIG_FILE_NAME+'''
    ; For Linux :   $XDG_CONFIG_HOME/'''+CONFIG_FILE_NAME+'''
    ;        or :   $HOME/'''+CONFIG_FILE_NAME+'''

    [general]
    output=text
    verbose=yes
    ssl_verify=no
    log_file=

    ; Use command line to setup authentication
    
    [esm]
    host=
    user=
    passwd=
    '''

POSSIBLE_FIELD_TYPES = [ 'BOOLEAN',
        'STRING',
        'CUSTOM',
        'INT2',
        'INT4',
        'INT8',
        'INT32',
        'INT64',
        'UINT8',
        'UINT16',
        'UINT32',
        'UINT64',
        'IPV4',
        'FLOAT',
        'SIGID',
        'SSTRING',
        'IPTYPE',
        'IP',
        'GUID',
        'MAC_ADDRESS',
        'LONG_CUSTOM',
        'HSTRING',
        'STRLIT',
        'AGG',
        'TIME4',
        'TIME8']

POSSIBLE_OPERATORS=['IN',
        'NOT_IN',
        'GREATER_THAN',
        'LESS_THAN',
        'GREATER_OR_EQUALS_THAN',
        'LESS_OR_EQUALS_THAN',
        'NUMERIC_EQUALS',
        'NUMERIC_NOT_EQUALS',
        'DOES_NOT_EQUAL',
        'EQUALS',
        'CONTAINS',
        'DOES_NOT_CONTAIN',
        'REGEX']

POSSIBLE_VALUE_TYPES=[
        {'type':'EsmWatchlistValue',    'key':'watchlist'},
        {'type':'EsmVariableValue',     'key':'variable'},
        {'type':'EsmBasicValue',        'key':'value'},
        {'type':'EsmCompoundValue',     'key':'values'}]

DEFAULT_TIME_RANGE="LAST_30_MINUTES"
POSSIBLE_TIME_RANGE=[
            "CUSTOM",
            "LAST_MINUTE",
            "LAST_10_MINUTES",
            "LAST_30_MINUTES",
            "LAST_HOUR",
            "CURRENT_DAY",
            "PREVIOUS_DAY",
            "LAST_24_HOURS",
            "LAST_2_DAYS",
            "LAST_3_DAYS",
            "CURRENT_WEEK",
            "PREVIOUS_WEEK",
            "CURRENT_MONTH",
            "PREVIOUS_MONTH",
            "CURRENT_QUARTER",
            "PREVIOUS_QUARTER",
            "CURRENT_YEAR",
            "PREVIOUS_YEAR"
        ]

POSSIBLE_ALARM_STATUS=[
        'acknowledged',
        'unacknowledged',
        '',
        None
    ]
POSSBILE_ROW_ORDER=[
        'ASCENDING',
        'DESCENDING'
]

DEFAULTS_EVENT_FIELDS=[
 "DSIDSigID",
 "SrcPort", 
 "AvgSeverity", 
 "DstPort", 
 "SrcIP", 
 "DstIP", 
 "SrcMac",
 "DstMac", 
 "LastTime"
]

ALARM_FILTER_FIELDS = ['id',
'summary',
'assignee',
'severity',
'triggeredDate',
'acknowledgedDate',
'acknowledgedUsername',
'alarmName']

"""
DETAILLED_ALARM_FIELDS =["filters",
"queryId",
"alretRateMin",
"alertRateCount",
"percentAbove",
"percentBelow",
"offsetMinutes",
"timeFilter",
"maximumConditionTriggerFrequency",
"useWatchlist",
"matchField",
"matchValue",
"healthMonStatus",
"assigneeId",
"escalatedDate",
"caseId",
"caseName",
"iocName",
"iocId",
"description",
"actions",
"events"]"""

ALARM_EVENT_FILTER_FIELDS=["eventId",
#"severity", ignored cause duplicated in ALARM_FILTER_FIELD
"ruleMessage",
"eventCount",
"sourceIp",
"destIp",
"protocol",
"lastTime",
"eventSubType"]




# -*- coding: utf-8 -*-
"""
    msiem session
"""
import ast
import re
import base64
import requests
import urllib3
import json
import getpass
import ssl
from urllib.parse import urljoin, unquote
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from abc import ABC
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

from .params import PARAMS
from .constants import BASE_URL, BASE_URL_PRIV, ASYNC_MAX_WORKERS, GENERAL_POST_TIMEOUT
from .exceptions import ESMException
from .utils import log, tob64
from .config import ESMConfig

class ESMSession(ABC):

    # static variables
    _headers={'Content-Type': 'application/json'}
    _executor = ThreadPoolExecutor(max_workers=ASYNC_MAX_WORKERS)
    _logged=False
    _config=None

    def __init__(self, conf_path=None, **config):

        #Config parsing
        ESMSession._config = ESMConfig(path=conf_path, **config)
        
        #singleton attributes mapping
        self._config=ESMSession._config
        self._executor = ESMSession._executor
        self._headers=ESMSession._headers
        self._logged=ESMSession._logged

    def __str__(self):
        return repr(self)

    def login(self):
        userb64 = tob64(self._config.get('esm', 'user'))
        passb64 = self._config.get('esm', 'passwd')
        
        resp = self.esmRequest('login', username=userb64, password=passb64, raw=True, secure=True)
        
        try :
            self._headers['Cookie'] = resp.headers.get('Set-Cookie')
            self._headers['X-Xsrf-Token'] = resp.headers.get('Xsrf-Token')

            self._logged = True

            if self.esmRequest('get_user_locale') == False :

                self._logged = False
                log.error("Failed to login ")
                raise ESMException("Login failed")

        except Exception as err :
            log.error("Failed to login \n"+str(err))
            raise ESMException("Login failed")

    @property
    def logged(self):
        return self._logged
        #return (not self.esmRequest('get_user_locale') == False)

    def logout(self):
        self.esmRequest('logout')
        self._logged=False

    def esmRequest(self, method, callback=None, raw=False, secure=False, asynch=True, **params):

        if not self.logged and method is not 'login' :
            self.login()

        method, data = PARAMS.get(method)

        if data :
            data =  data % params
            data = ast.literal_eval(''.join(data.split()))
        
        if method :
            method = method % params

        post=dict()

        try :
            post = self._post(method, data, callback, raw, secure, asynch)
            return post

        except Exception as err:
            log.error(str(err))
            return False

    def _post(self, method=None, data=None, callback=None, raw=False, secure=False, asynch=True):
        """
        Helper method
        If method is all upper cases, a private API is done.
        Private API is under /ess/ and public api is under /rs/esm
        """

        url=str()
        privateApiCall=False
        result=None

        #Handling private API calls formatting
        if method == method.upper():
            privateApiCall=True
            url = BASE_URL_PRIV
            if data :
                data = self._format_params(method, **data)
        
        #Normal API calls
        else:
            url = BASE_URL
            if data:
                data = json.dumps(data)

        if data and not secure :
            log.debug('DATA : '+data)

        try :
            if asynch :
                future = self._executor.submit(
                    requests.post, 
                    urljoin(url.format(self._config.get('esm', 'host')), method),
                    data=data, 
                    headers=ESMSession._headers,
                    verify=self._config.getboolean('general','ssl_verify'),
                    timeout=GENERAL_POST_TIMEOUT
                    )

                result = future.result()

            elif not asynch :
                result = requests.post(
                    urljoin(url.format(self._config.get('esm', 'host')), method),
                    data=data, 
                    headers=ESMSession._headers,
                    verify=self._config.getboolean('general','ssl_verify'),
                    timeout=GENERAL_POST_TIMEOUT
                )

            if not secure :
                pass
                #log.debug('RESULT : '+result.text)

            if raw :
                return result

            else:
                try:
                    result.raise_for_status()

                except requests.HTTPError as e :
                    log.error(str(e)+' '+str(result.text))

                else: #
                    try:
                        result = result.json()
                        result = result.get('return')

                    except json.decoder.JSONDecodeError:
                        result = result.text

                    if privateApiCall :
                        result = self._format_priv_resp(result)

                    if callback:
                        result = callback(result)

                    return result

        except ConnectionError as e:
            log.critical(e)
            raise
        except requests.exceptions.Timeout as e:
            log.error(e)
            pass
        except requests.exceptions.TooManyRedirects as e :
            log.error(e)
            pass
        except Exception as e:
            log.error(e)
        
    @staticmethod
    def _format_params(cmd, **params):
        """
        Format private API call.
        From mfe_saw project at https://github.com/andywalden/mfe_saw
        """
        params = {k: v for k, v in params.items() if v is not None}
        params = '%14'.join([k + '%13' + v + '%13' for (k, v) in params.items()])
        
        if params:
            params = 'Request=API%13' + cmd + '%13%14' + params + '%14'
        else:
            params = 'Request=API%13' + cmd + '%13%14'
        return params

    @staticmethod
    def _format_priv_resp(resp):
        """
        Format response from private API
        From mfe_saw project at https://github.com/andywalden/mfe_saw
        """
        resp = re.search('Response=(.*)', resp).group(1)
        resp = resp.replace('%14', ' ')
        pairs = resp.split()
        formatted = {}
        for pair in pairs:
            pair = pair.replace('%13', ' ')
            pair = pair.split()
            key = pair[0]
            if key == 'ITEMS':
                value = pair[-1]
            else:
                value = unquote(pair[-1])
            formatted[key] = value
        return formatted


# -*- coding: utf-8 -*-
"""
    msiem config
"""

from configparser import ConfigParser, NoSectionError, MissingSectionHeaderError
import os
import sys
from getpass import getpass
try:
    from .utils import log, tob64
    from .constants import DEFAULT_CONFIG_FILE, CONFIG_FILE_NAME
except ImportError :
    from utils import log, tob64
    from constants import DEFAULT_CONFIG_FILE, CONFIG_FILE_NAME

class ESMConfig(ConfigParser):
    """
    Find the config settings which include:
    
    [general]
    output=<text/json>
    verbose=<yes/no>
    ssl_verify=<yes/no>
    log_file=<path>

    [esm]
    host=<fqdn>
    user=<username>
    passwd=<b64password>
    
    
    ESMConfig(esm={'host':'10.2.2.1', 'user':'tristan'})
    """

    def __init__(self, path=None, **kwconfig):
        """
        Initialize the Config instance.
        """
        super().__init__()

        self._path=None

        if not path :

            path = self._find_ini_location()

            try :
                f=self.read(path)
                if len(f) == 0:
                    raise FileNotFoundError
            except :
                log.warning("Config file inexistant or currupted, applying defaults")
                self.read_string(DEFAULT_CONFIG_FILE)

            else :
                self._path = path

        if kwconfig :
            log.info("Applying config from manual specifications")
            for section in kwconfig :
                try :
                    if self.has_section(section):
                        for key in kwconfig[section] :
                            log.info("Setting ["+section+"] "+key+" = "+kwconfig[section][key])
                            self.set(section, key, kwconfig[section][key])
                    else:
                        log.warning("Skipping unknown config section "+section)

                    self.write()

                except Exception as err:
                    log.error("An error occured when reading the manual config dict. \n"+str(err))
                    raise
        
    def write(self, path=None):
        if not path :
            path=self._path
            if not path :
                path = self._find_ini_location()
        
        if path :
            try :
                if not os.path.exists(os.path.dirname(path)):
                    os.makedirs(os.path.dirname(path))

                with open(path, 'w') as conf:
                    super().write(conf)

                self._path=path
                log.info("Config file wrote at "+path)

            except Exception as err :
                log.error("An error occured when writing the config file. \n"+str(err))
        else :
            log.error("Couldn't write config file, no specified path.")

    @staticmethod
    def _find_ini_location():
        """
        Attempt to locate the conf.ini file 
        """
        conf_path=None

        if 'APPDATA' in os.environ:
            conf_path = os.environ['APPDATA']

        elif 'XDG_CONFIG_HOME' in os.environ:  
            conf_path = os.environ['XDG_CONFIG_HOME']

        elif 'HOME' in os.environ:  
            conf_path = os.path.join(os.environ['HOME'])
    
       
        if conf_path  :
            conf_path=(os.path.join(conf_path, CONFIG_FILE_NAME))
            #log.info("Assuming config file at {}".format(conf_path))
        else :
            raise SystemError("Couldn't find either APPDATA, XDG_CONFIG_HOME or HOME environemnt variable to locate the conf.ini file.")

        return(conf_path)

    def setAuth(self):
        self.set('esm', 'host', input('Please enter the ESM FQDN: '))
        self.set('esm', 'user', input('Please enter your username: '))
        self.set('esm', 'passwd', tob64(getpass()))


# -*- coding: utf-8 -*-
"""
    msiem command
"""

"""

msiem 
    --help #Exits after
    --version #Exists after
    --batch

msiem config
    --help
    --list/-l
    --set auth
    --set <section.key>=<value>

msiem esm 
    --help/-h
    --time/-t
    --version/-v
    --tz/--timezones
    --disks/-d
    --callhome
    --ram
    --status
    --backup-status
    --buildstamp
    --recs
    --rule

    #Todo
    --nslookup
    --fields -> display list of possible fields with qryGetSelectFields
    --filters -> display list of possible filters with qryGetFilterFields

    #FullBackUp
    --backup esm_full_backup
    https://github.com/andywalden/esm_full_backup/

msiem ds
    --help/-h
    --add/-a interactive si manque arg
        name=DC01_DNS 
        ip=10.10.1.34 
        rec_ip=172.16.15.10
        type=linux
        Will add a new datasource and also call --info
        Ask for confirmation unless --force

    --del/-d <id or name> if resolves to only one device
        Ask for confirmation unless --force

    --search/-s <pattern>
    --searchgroup/-g
    --info/-i <id or name>
    --times/-t
    (--refresh)
    --recs --> ?
    --ipvalid <>
    --addclient <ds> <> ...

    --checkinactive
    Queries a McAfee ESM for inactive data sources. from https://github.com/andywalden/esmcheckds2

msiem alarms

    --search/-s [default]

    --ack/-a
        Ask for confirmation unless --force

    --del/-d
        Ask for confirmation unless --force

    --unack/-u

    --filter/-f    
                -f event.srcip 10.2.2.2
                -f event.dstip 10.2.2.1
                -f event.msg=http-vulnerability
                -f event.name=Bob le bricoleur'
                #and others under event. #See documentation

                -f summary=<>
                -f severity=90
                -f ackuser=tristan
                -f assignee=tristan
                -f name=Critical
                -f acked=[yes/no/all]

                -f ackdate.begins=2019-03-25
                -f ackdate.ends=2019-03-25
                -f time=2019-03-25T14:00
                -f window=1h
                -f timerange=LAST_24_HOURS
                -f begins=2019-03-25
                vends=2019-03-25

                #No filtering on DetailedAlarm fields value !
                #Only a couple basic event related fields are available for filtering
                #Filters are ANDed together

    --info/-i <id>
        --> show the conditionnal xml tree and alarm meta data

    --force


msiem events
    --search

    --filter
            -f sourceip=10.2.2.2
            -f destinationip=10.2.2.1
            -f message=http-vulnerability
            -f 'name=Bob le bricoleur'

            -f time=2019-03-25T14:00
            -f window=1h
            -f timerange=LAST_24_HOURS
            -f begins=2019-03-25
            -f ends=2019-03-25

            #and others under event. #See documentation
            #Filters are ANDed together

    --fields
        #Fields to add to the defaults settings

    --order <order>

msiem watchlist
    Inteface to whatchlist operations

msiem service
    Opens a service now mcfee ticket.
    Callable by the Execute function of the ESM.
    https://github.com/andywalden/mfe2snow

msiem syslog
    Generate syslog message on the SIEM

msiem case
    --close https://github.com/andywalden/esm_close_cases
    --open <alarm>
    --msg <message>

msiem elm 
    Export ELM files back to original format
    https://github.com/andywalden/elmex/blob/master/elmex.sh
    
"""

import argparse
try:
    from .config import ESMConfig
except ImportError :
    from config import ESMConfig

def parseArgs():
    parser = argparse.ArgumentParser(description='McAfee SIEM Command Line Interface and Python API',
                usage='Use "msiem --help" for more information',
                formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('--version')
    parser.add_argument('--batch')

    commands = parser.add_subparsers()

    config_command = commands.add_parser('config')
    config_command.set_defaults(func=config)
    config_command.add_argument('--list')
    config_command.add_argument('--set')

    return (parser.parse_args())

def config(args):
    c=ESMConfig()
    if args.set is not None :
        if 'auth' in args.set :
            c.setAuth()
            c.write()

def main():
    args = parseArgs()
    args.func(args)

if __name__ == "__main__":
    main()






# -*- coding: utf-8 -*-
"""
    msiem utils
"""

import time
import base64
from functools import wraps
import logging
from .constants import POSSIBLE_TIME_RANGE
from .exceptions import ESMException
from datetime import datetime, timedelta

def getLog(v=True, logfile=None):

    log = logging.getLogger()
    log.setLevel(logging.DEBUG)

    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

    std = logging.StreamHandler()
    std.setLevel(logging.DEBUG)
    std.setFormatter(formatter)

    if v :
        std.setLevel(logging.DEBUG)
    else :
        std.setLevel(logging.INFO)
        
    log.addHandler(std)

    if logfile :
        fh = logging.FileHandler(logfile)
        fh.setLevel(logging.INFO)
        fh.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
        log.addHandler(fh)

    return (log)

log = getLog()

def dehexify(data):
    """
    A URL and Hexadecimal Decoding Library.

    Credit: Larry Dewey
    """

    hexen = {
        '\x1c': ',',  # Replacing Device Control 1 with a comma.
        '\x11': '\n',  # Replacing Device Control 2 with a new line.
        '\x12': ' ',  # Space
        '\x22': '"',  # Double Quotes
        '\x23': '#',  # Number Symbol
        '\x27': '\'',  # Single Quote
        '\x28': '(',  # Open Parenthesis
        '\x29': ')',  # Close Parenthesis
        '\x2b': '+',  # Plus Symbol
        '\x2d': '-',  # Hyphen Symbol
        '\x2e': '.',  # Period, dot, or full stop.
        '\x2f': '/',  # Forward Slash or divide symbol.
        '\x7c': '|',  # Vertical bar or pipe.
    }

    uri = {
        '%11': ',',  # Replacing Device Control 1 with a comma.
        '%12': '\n',  # Replacing Device Control 2 with a new line.
        '%20': ' ',  # Space
        '%22': '"',  # Double Quotes
        '%23': '#',  # Number Symbol
        '%27': '\'',  # Single Quote
        '%28': '(',  # Open Parenthesis
        '%29': ')',  # Close Parenthesis
        '%2B': '+',  # Plus Symbol
        '%2D': '-',  # Hyphen Symbol
        '%2E': '.',  # Period, dot, or full stop.
        '%2F': '/',  # Forward Slash or divide symbol.
        '%3A': ':',  # Colon
        '%7C': '|',  # Vertical bar or pipe.
    }

    for (enc, dec) in hexen.items():
        data = data.replace(enc, dec)

    for (enc, dec) in uri.items():
        data = data.replace(enc, dec)

    return data


def timethis(func):
    """
    Decorator that reports the execution time.
    Credit: andywalen
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        """Wrapper"""
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(func.__name__, end-start)
        return result
    return wrapper

def tob64(s):
    if type(s) is str:
        return base64.b64encode(s.encode('utf-8')).decode()
    log.warning("Trying to convert non string object to b64")

def fromb64(s):
    if type(s) is str:
        return base64.b64decode(s.encode('utf-8')).encode()
    log.warning("Trying to convert non string object from b64")

def getTimes(timeFrame):
    t=timeFrame
    now=datetime.now()
    times=tuple()

    if t is 'LAST_MINUTE' :
        times=(now-timedelta(seconds=60), now)
        
    elif t is 'LAST_10_MINUTES':
        times=(now-timedelta(minutes=10), now)

    elif t is 'LAST_30_MINUTES':
        times=(now-timedelta(minutes=30), now)

    elif t is 'LAST_HOUR':
        times=(now-timedelta(minutes=60), now)

    elif t is 'CURRENT_DAY':
        times=(now.replace(hour=0, minute=0, second=0), now.replace(hour=24, minute=59, second=59))

    elif t is 'PREVIOUS_DAY':
        yesterday=now-timedelta(hours=24)
        times=(yesterday.replace(hour=0, minute=0, second=0), yesterday.replace(hour=24, minute=59, second=59))

    elif t is 'LAST_24_HOURS':
        times=(now-timedelta(hours=24), now)

    elif t is 'LAST_2_DAYS':
        times=(now-timedelta(days=2), now)

    elif t is 'LAST_3_DAYS':
        times=(now-timedelta(days=3), now)

    else :
        raise ESMException("Timerange "+t+" is not supported for custom cumputation")
    
    return(times[0].isoformat(), times[1].isoformat())
    
    """
    elif t is 'CURRENT_WEEK':
        pass
    elif t is 'PREVIOUS_WEEK':
        pass
    elif t is 'CURRENT_MONTH':
        pass
    elif t is 'PREVIOUS_MONTH':
        pass
    elif t is 'CURRENT_QUARTER':
        pass
    elif t is 'PREVIOUS_QUARTER':
        pass
    elif t is 'CURRENT_YEAR':
        pass
    elif t is 'PREVIOUS_YEAR':
        pass"""



# -*- coding: utf-8 -*-
"""
    msiem query
"""
import json
import time
from .session import ESMSession
from .exceptions import ESMException
from .utils import log, getTimes
from .constants import (POSSIBLE_TIME_RANGE,
    POSSIBLE_FIELD_TYPES, 
    POSSIBLE_OPERATORS, 
    POSSIBLE_VALUE_TYPES,
    POSSIBLE_ALARM_STATUS,
    POSSBILE_ROW_ORDER,
    DEFAULTS_EVENT_FIELDS)

class QueryBase(ESMSession):
    def __init__(self, time_range=None, start_time=None, end_time=None):
        super().__init__()

        #Declaring attributes
        self._time_range=str()
        self._start_time=str()
        self._end_time=str()
        
        #init attributes
        if time_range and time_range is not 'CUSTOM':
            self.time_range=time_range

        elif start_time and end_time :
            self.time_range='CUSTOM'
            self.start_time=start_time
            self.end_time=end_time

        else :
            raise ESMException("The query must have valid time specifications. Please refer to documentation.")

    @property
    def time_range(self):
        return self._time_range

    @property
    def start_time(self):
        return self._start_time

    @property
    def end_time(self):
        return self._end_time

    @time_range.setter
    def time_range(self, time_range):
        """
        Set the time range of the query to the specified string value. Default : LAST_3_DAYS
        """

        try:
            if time_range in POSSIBLE_TIME_RANGE :
                self._time_range=time_range
                
            else:
                raise ESMException("The time range must be in "+str(POSSIBLE_TIME_RANGE))

        except:raise
    
    @start_time.setter
    def start_time(self, start_time):
        """
        Set the time start of the query. Time range must be CUSTOM for this to work.
        """
        try:
            if start_time is not None :
                if self._time_range == 'CUSTOM':
                    self._start_time=start_time
                    
                else:
                    raise ESMException("The time range must be 'CUSTOM' if you want to specify a custom start time")
            else:
                self._start_time=None
                
        except:raise
    
    @end_time.setter
    def end_time(self, end_time):
        """
        Set the time end of the query. Time range must be CUSTOM for this to work.
        """
        try:
            if end_time is not None :
                if self._time_range == 'CUSTOM':
                    self._end_time=end_time
                   
                else:
                    raise ESMException("The time range must be 'CUSTOM' if you want to specify a custom end time")
            else :
                self._end_time=None
                
        except:raise

    def execute(self):
        raise ESMException("Base query can't be executed.")

class AlarmQuery(QueryBase):

    def __init__(self, status='', page_size=5000, page_number=1, filters=None, **args):

        """
        filters : [(field, [values]), (field, [values])]
        field can be an EsmTriggeredAlarm field, EsmTriggeredAlarmDetail or an EsmTriggeredAlarmEvent field
        """

        super().__init__(**args)

        #Declaring attributes
        self._status=str()
        self._page_size=int()
        self._page_number=int()

        self._filters = list(tuple())

        #Setting attributes
        self.status=status
        self.page_size=page_size
        self.page_number=page_number

    @property
    def status(self):
        return self._status

    @property
    def page_size(self):
        return self._page_size

    @property
    def page_number(self):
        return self._page_number

    @status.setter
    def status(self, status):
        """
        Set the status filter of the alarm query. 'acknowledged', 'unacknowledged', '' or null -> all (default is '')
        """
        try:
            if type(status) is str : 
                if status.lower() in POSSIBLE_ALARM_STATUS :
                    self._status=status
                    
                else:
                    raise ESMException("Illegal value of status filter. The status must be in "+str(POSSIBLE_ALARM_STATUS))

        except:raise
    
    @page_size.setter
    def page_size(self, page_size):
        """
        Set the page size, the number of alarms to return per page (default is 5000, max is 5000).
        """
        try:
            if page_size > 0 and page_size <= 5000 :
                self._page_size=page_size
                
            else:
                raise ESMException("The page size must be in 0-5000")

        except:raise

    @page_number.setter
    def page_number(self, page_number):
        """
        Set the page number. Which page of alarms we want to return (default is 1)
        """
        self._page_number=page_number
        
    def execute(self):
        """"
        Execute the query.
        Returns a list of Alarms.
        """

        resp=None
        if self.time_range is 'CUSTOM' :
            resp=self.esmRequest(
                'get_alarms_custom_time',
                time_range=self.time_range,
                start_time=self.start_time,
                end_time=self.end_time,
                status=self.status,
                page_size=self.page_size,
                page_number=self.page_number
                )

        else :
            resp=self.esmRequest(
                'get_alarms',
                time_range=self.time_range,
                status=self.status,
                page_size=self.page_size,
                page_number=self.page_number
                )
            

        alarms=list()
        for alarm_data in resp :
            alarm = Alarm(**alarm_data)
            alarms.append(alarm)

        return alarms

class Alarm(ESMSession):
    """
    Alarm object. 
    Gives possibility to access alarm fields and related events.
    Delete, Ack, or Unack alarm.
    """

    def __init__(self, **args):
        
        self.id = {"value" : 0}
        self.summary = ''
        self.assignee = ''
        self.severity = 0
        self.triggeredDate = ''
        self.acknowledgedDate = ''
        self.acknowledgedUsername = ''
        self.alarmName = ''
        self.conditionType = 0

        self.__dict__.update(args)

        self._detailed = None

        try :
            if self.id['value'] == 0:
                raise ESMException("The id of the alarm haven't been initialized ! Make sure you pass a dict object with a 'value' field as the id.")
        except KeyError :
            raise KeyError("Make sure you pass a dict object with a 'value' field as the id.")
        except ESMException:
            raise
        
        super().__init__()

    @property
    def detailed(self):
        if self._detailed is None :
            self._detailed = DetailedAlarm(self)
        return self._detailed

    @property
    def events(self):
        return self.detailed.events

    def delele(self):
        self.esmRequest('delete_alarms', ids=[self.id])
        return
    
    def acknowledge(self):
        self.esmRequest('ack_alarms', ids=[self.id])
        return

    def unacknowledge(self):
        self.esmRequest('unack_alarms', ids=[self.id])
        return  

class DetailedAlarm(Alarm):
    """
    Alarm details object. Based on EsmTriggeredAlarmDetail
    """
    def __init__(self, alarm):
        
        """
        self.id = {"value" : 0} #duplicate
        self.summary = '' #duplicate
        self.assignee = '' #duplicate
        self.severity = 0 #duplicate
        self.triggeredDate  = '' #duplicate
        self.acknowledgedDate  = ''#duplicate
        self.acknowledgedUsername  = ''#duplicate
        self.alarmName  = ''#duplicate
        self.conditionType  = 0 #duplicate
        """

        super().__init__(id=alarm.id)
        
        """
        self.filters  = ''
        self.queryId = 0
        self.alretRateMin = 0
        self.alertRateCount = 0
        self.percentAbove = 0
        self.percentBelow = 0
        self.offsetMinutes = 0
        self.timeFilter  = ''
        self.maximumConditionTriggerFrequency = 0
        self.useWatchlist = ''
        self.matchField = ''
        self.matchValue = ''
        self.healthMonStatus = ''
        self.assigneeId = 0
        self.escalatedDate = ''
        self.caseId = 0
        self.caseName = ''
        self.iocName = 0
        self.iocId = 0
        self.description = ''
        self.actions = ''
        """

        self._events = list()

        resp = self.esmRequest('get_alarm_details', id=alarm.id)
        
        #self.__dict__.update(resp)

        self._events = resp['events']

    @property
    def detailed(self):
        return self

    @property
    def events(self):
        return self._events

class Event(ESMSession):
    """ Based on EsmTriggeredAlarmEvent """
    def __init__(self, **args):
        self.fields=dict().update(**args)

class EventQuery(QueryBase):
    """
    Interface to qryExecuteDetail?type=EVENT method.
    """

    #Declaring static value
    _possible_fields = None

    def __init__(self, fields=None, filters=None, 
        limit=5000, offset=0, order=None, compute_time_range=True, **args):
        """
       
        fields list of string, need to convert that into list of {
            "name": "(name)",
            "typeBits": 0,
            "id": "(id)"
            }

        order tuple (direction, field) to list of {
            "direction": "ASCENDING or DESCENDING",
            "field": {
                "name": ""
                }
            }

        filters 
            1 / list of {
                (field , [values]) #easy mode
            }

            2 / list of QueryFilter
        """

        self._compute_time_range = compute_time_range
        
        super().__init__(**args)\

        #Singleton attribute mapping
        self._possible_fields = EventQuery._possible_fields

        #Constant
        self._type='EVENT'
        self._groupType='NO_GROUP'

        if self._possible_fields is None :
            self._possible_fields = self.esmRequest('get_possible_fields', type=self._type, groupType=self._groupType)

        #Declaring attributes
        self._fields=set()
        self._filters=list()

        self._limit=int()
        self._offset=int()

        self._reverse=bool()

        self._order=[{
            "direction": None,
            "field": {
                "name": None
                }
            }]

        self.fields = DEFAULTS_EVENT_FIELDS
        self.fields = fields

        self.filters = filters

        self.order = order
        
        self._limit = limit
        self._offset = offset

        self._query = None

        self._executed = False

    @property
    def fields(self):
        return([{'name':value} for value in list(self._fields)])

    @property
    def filters(self):
        return([f.configDict() for f in self._filters])

    @property
    def limit(self):
        return(self._limit)

    @property
    def offset(self):
        return(self._offset)

    @property
    def reverse(self):
        return(self._reverse)

    @property
    def order(self) :       
        return(self._order)

    @property
    def time_range(self):
        return(super().time_range)

    @order.setter
    def order(self, order):
        #tuple (direction, field)
        #TODO check valid value
        if type(order) is tuple :
            self._order[0]['direction']=order[0]
            self._order[0]['field']['name']=order[1]

    @filters.setter
    def filters(self, filters):
        for values in filters :
            self.add_filter(values)

    def add_filter(self, fil):
        if type(fil) is tuple :
            self._filters.append(FieldFilter(*fil))

        elif isinstance(fil, QueryFilter) :
            self._filters.append(fil)
        
        else :
            raise ESMException("Sorry the filters must be either a tuple(fiels, [values]) or a QueryFilter sub class.")

    @fields.setter
    def fields(self, fields):
        if fields :
            for f in fields :
                if f :
                    self.add_field(f) 

    def add_field(self, field):
        try:
            if any(f.get('name', None) == field for f in self._possible_fields):
                self._fields.add(field)
            else:
                raise ESMException("Illegal value for the value "+field+". The field must be in "+ str([f['name'] for f in self._possible_fields]))
        except GeneratorExit :
            pass
        except:raise

    @time_range.setter
    def time_range(self, time_range):
        """
        Set the time range of the query to the specified string value.
        """
        if time_range is not 'CUSTOM' and self._compute_time_range :
            try :
                times = getTimes(time_range)
                self._time_range='CUSTOM'
                self._start_time=times[0]
                self._end_time=times[1]

            except ESMException :
                log.warning('The choosen time range is not fully supported. This mean that only the first page of events will be returned. '+str(ESMException))
                self._time_range=time_range
        else :
            self._time_range=time_range
  

    def execute(self):
        """"
            Execute the query.
            Returns a list of Events.
        """

        if self.time_range is 'CUSTOM' :
            self._query=self.esmRequest(
                'event_query_custom_time',
                asynch=False,

                time_range=self.time_range,
                start_time=self.start_time,
                end_time=self.end_time,
                #order=self.order, TODO support order
                fields=self.fields,
                filters=self.filters,
                limit=self.limit,
                offset=self.offset,
                includeTotal=False
                )

        else :
            self._query=self.esmRequest(
                'event_query',
                asynch=False,

                time_range=self.time_range,
                #order=self.order, TODO support order
                fields=self.fields,
                filters=self.filters,
                limit=self.limit,
                offset=self.offset,
                includeTotal=False
                )
        
        log.debug("EsmRunningQuery object : "+str(self._query))

        if self._waitFor(self._query['resultID']):
            self._executed = True

        return (self.getEvents())

    def _waitFor(self, resultID):
        log.debug("Waiting for the query to be executed on the SIEM...")
        while True:
            status = self.esmRequest('query_status', resultID=resultID)
            if status : 
                if status['complete'] is True :
                    return True
                else :
                    time.sleep(0.5)
            else :
                raise ESMException("There was an error while waiting for the query to execute.")

    def getEvents(self, startPos=0, numRows=None):
        if self._executed :
            if not numRows :
                numRows=self.limit
                
            result=self.esmRequest('query_result', startPos=startPos, numRows=numRows, resultID=self._query['resultID'])
            events=self._parseResult(result['columns'], result['rows'])
            return events
        else :
            raise ESMException("Query need to be executed before to be able to get events.")

    def _parseResult(self, columns, rows):
        events=list()
        for row in rows :
            event=dict()
            for i in range(len(columns)-1):
                event.update({columns[i]['name']:row['values'][i]})
            events.append(event)
        return events

class QueryFilter(ESMSession):

    _possible_filters = None

    def __init__(self):
        super().__init__()

        #Mapping singleton attributes
        self.__possible_filters = QueryFilter._possible_filters

        #Setting up static constant
        if self._possible_filters is None :
            self._possible_filters = self._get_possible_filters()

    def _get_possible_filters(self):
        return(self.esmRequest('get_possible_filters'))

    def configDict(self):
        return 'TODO'

class GroupFilter(QueryFilter):
    """
        Based on EsmFilterGroup
    """
    def __init__(self, *filters, logic='AND'):
        super().__init__()

        #Declaring attributes
        self._filters=filters
        self._logic=logic

    def configDict(self):
        return({
            "type": "EsmFilterGroup",
            "filters": [f.configDict() for f in self._filters],
            "logic":self._logic
            })
        
class FieldFilter(QueryFilter):
    """
    Based on EsmFieldFilter
    """

    def __init__(self, name, values, operator='IN'):
        super().__init__()

        #Declaring attributes
        self._name=str()
        self._operator=str()
        self._values=list()

        self._name = name
        self.operator = operator
        self.values = values

    def configDict(self):
        return ({
            "type": "EsmFieldFilter",
            "field": {"name": self._name},
            "operator": self._operator,
            "values": self._values
            })

    @property
    def field_name(self):
        return (self._field_name)
    
    @property
    def operator(self):
        return (self._operator)

    @property
    def values(self):
        return (self._values)

    @field_name.setter
    def field_name(self, name):
        try:
            if any(f.get('name', None) == name for f in self._possible_filters):
                self._field_name = name
            else:
                raise ESMException("Illegal value for the "+name+" field. The filter must be in :"+str([f['name'] for f in self._possible_filters]))
        except GeneratorExit:
            pass
        except:
            raise

    @operator.setter
    def operator(self, operator):
        try:
            if operator in POSSIBLE_OPERATORS :
                self._operator = operator
            else:
                raise ESMException("Illegal value for the filter operator "+operator+". The operator must be in "+str(POSSIBLE_OPERATORS))
        except:
            raise
        
    def add_value(self, type, **args):
        """
        Please refer to the EsmFilterValue documentation
        """
        try:
            type_template=None
            for possible_value_type in POSSIBLE_VALUE_TYPES :
                if possible_value_type['type'] == type :
                    type_template=possible_value_type
                    break

            if type_template is not None :
                if type_template['key'] in args :
                    self._values.append({'type':type, type_template['key']:args[type_template['key']]})

                else:
                    raise ESMException("You must provide a valid named parameter containing the value(s). The key must be in "+str(POSSIBLE_VALUE_TYPES))

            else:
                raise ESMException("Illegal value type for the value filter. The type must be in "+str(POSSIBLE_VALUE_TYPES))
        except:
            raise

    def add_basic_value(self, value):
        self.add_value(type='EsmBasicValue', value=value)

    @values.setter
    def values(self, values):
        for v in values :
            if type(v) is dict :
                self.add_value(**v)

            elif type(v) is str :
                self.add_basic_value(v)






try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='msiem',
    url='https://github.com/tristanlatr/msiem',
    version='0.0.1',
    packages=['msiem',],
    entry_points = {
        'console_scripts': ['msiem=msiem.command:main'],
    },
    install_requires=[
          'requests',
    ],
    license='The MIT License',
    long_description=open('README.md').read(),
    test_suite="tests"
)